/*
 *  serial.c -- Machine/OS dependent serial I/O 
 *
 *  Copyright 1997 (unix code)  John E. Stone (j.stone@acm.org)
 *
 *  $Id: serial.c,v 1.1 2007/09/02 06:42:04 johns Exp $
 *
 */

#if !defined(WIN32) && !defined(_MSC_VER)

/*
 * Unix serial I/O code by John Stone
 */

#if 0
#define _POSIX_SOURCE 1
#endif

#include <stdio.h>
#include <termios.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include "serial.h" /* protos and types for this file */

typedef struct {
  int fd;        /* serial port device file descriptor */
} commstruct;

int rs232_comm_open(const char * portname, RS232CommHandle * commhandleptr, int baud) {
  struct termios rs232termio;
  commstruct * comm;
  int baudrate;

  switch (baud) {
    case     50: baudrate =     B50; break;
    case     75: baudrate =     B75; break;
    case    110: baudrate =    B110; break;
    case    134: baudrate =    B134; break;
    case    150: baudrate =    B150; break;
    case    200: baudrate =    B200; break;
    case    300: baudrate =    B300; break;
    case    600: baudrate =    B600; break;
    case   1200: baudrate =   B1200; break;
    case   1800: baudrate =   B1800; break;
    case   2400: baudrate =   B2400; break;
    case   4800: baudrate =   B4800; break;
    case   9600: baudrate =   B9600; break;
    case  19200: baudrate =  B19200; break;
    case  38400: baudrate =  B38400; break;
    case  57600: baudrate =  B57600; break;
    case 115200: baudrate = B115200; break;
    case     0:
       default:
      return -1; /* failed to set requested baud rate */
  }

  *commhandleptr = NULL;

  comm = malloc(sizeof(commstruct));
  if (comm == NULL)
    return -1;

  comm->fd = open(portname, O_RDWR | O_NONBLOCK | O_NOCTTY);

  if (comm->fd == -1) {
    free(comm);
    return -1;  /* failed open of comm port */
  }
  tcgetattr(comm->fd, &rs232termio);

#if 0
  /* TIOCEXCL exclusive access by this process */
#if defined(TIOCEXCL)
  if (ioctl(comm->fd, TIOCEXCL) < 0) {
    return -1; /* couldn't get exclusive use of port */
  }
#endif
#endif

  rs232termio.c_lflag = 0;
  rs232termio.c_lflag = 0;
  rs232termio.c_iflag = IGNBRK | IGNPAR;
  rs232termio.c_oflag = 0;
  rs232termio.c_cflag = CREAD | CS8 | CLOCAL | HUPCL;
  rs232termio.c_cc[VEOL] = '\r';
  rs232termio.c_cc[VERASE] = 0;
  rs232termio.c_cc[VKILL] = 0;
  rs232termio.c_cc[VMIN] = 0;
  rs232termio.c_cc[VTIME] = 0;

  /* use of baud rate in cflag is deprecated according to the */
  /* single unix spec, also doesn't work in IRIX > 6.2        */
  cfsetispeed(&rs232termio, baudrate);
  cfsetospeed(&rs232termio, baudrate);

  tcsetattr(comm->fd, TCSAFLUSH, &rs232termio);

  *commhandleptr = (RS232CommHandle) comm;
 
  return 0;
}

int rs232_comm_write(RS232CommHandle commhandle, const char * buf) {
  commstruct * comm = (commstruct *) commhandle;

  if (comm == NULL) 
    return -1;

  return write(comm->fd, buf, strlen(buf));
}

int rs232_comm_read(RS232CommHandle commhandle, char * buf, int sz) {
  commstruct * comm = (commstruct *) commhandle;

  if (comm == NULL) 
    return -1;

  return read(comm->fd, buf, sz);
}

int rs232_comm_close(RS232CommHandle * commhandleptr) {
  commstruct * comm = (commstruct *) (*commhandleptr);

  if (comm == NULL)
    return -1;

  close(comm->fd);

  free(*commhandleptr);
  *commhandleptr = NULL;

  return 0;
}

int rs232_comm_get_status(RS232CommHandle commhandle, int *dsr, int *dtr, int *rts, int *cts, int *cd, int *rng) {
  int status;
  commstruct * comm = (commstruct *) commhandle;

  if (comm == NULL)
    return -1;

  if (ioctl(comm->fd, TIOCMGET, &status) == -1)
    return -1;

  if (dsr != NULL) 
    *dsr = ((status & TIOCM_DSR) != 0);

  if (dtr != NULL) 
    *dtr = ((status & TIOCM_DTR) != 0);

  if (rts != NULL) 
    *rts = ((status & TIOCM_RTS) != 0);

  if (cts != NULL) 
    *cts = ((status & TIOCM_RTS) != 0);

  if (cd != NULL) 
    *cd = ((status & TIOCM_CD) != 0);

  if (rng != NULL) 
    *rng = ((status & TIOCM_RNG) != 0);

  return 0;
}


int rs232_comm_set_rts(RS232CommHandle commhandle, int level) {
  int status;
  commstruct * comm = (commstruct *) commhandle;

  if (comm == NULL)
    return -1;

  if (ioctl(comm->fd, TIOCMGET, &status) == -1)
    return -1;
  if (level)
    status |= TIOCM_RTS;
  else
    status &= ~TIOCM_RTS;
  if (ioctl(comm->fd, TIOCMSET, &status) == -1)
    return -1;
  return 0;
}


int rs232_comm_set_dtr(RS232CommHandle commhandle, int level) {
  int status;
  commstruct * comm = (commstruct *) commhandle;

  if (comm == NULL)
    return -1;

  if (ioctl(comm->fd, TIOCMGET, &status) == -1)
    return -1;
  if (level)
    status |= TIOCM_DTR;
  else
    status &= ~TIOCM_DTR;
  if (ioctl(comm->fd, TIOCMSET, &status) == -1)
    return -1;
  return 0;
}

/* end of unix code */

#else 

/* Win32 code */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <windows.h>
#include "serial.h" /* protos and types for this file */

typedef struct {
  HANDLE hCom; /* communication handle */
} commstruct;

int rs232_comm_open(const char * portname, RS232CommHandle * commhandleptr, int baud) {
  commstruct * comm;
  DCB dcb;
  COMMTIMEOUTS cto;
  HANDLE hCom;
  BOOL fSuccess;

  hCom = CreateFile( portname,
                     GENERIC_READ | GENERIC_WRITE,
                     0,    // comm devices must be opened w/exclusive-access
                     NULL, // no security attributes
                     OPEN_EXISTING, // comm devices must use OPEN_EXISTING
                     0, // not overlapped I/O
                     NULL);  // hTemplate must be NULL for comm devices     );

  if (hCom == INVALID_HANDLE_VALUE) {
    perror("rs232_comm_open: cannot open serial port");
    return -1;
  }

  if (!(fSuccess = GetCommState(hCom, &dcb))) {
    perror("rs232_comm_open: cannot get serial port configuration settings");
    CloseHandle(hCom);
    return -1;
  }

  // Enable the hardware data-ready line (the TNG-3 uses this for power).
  dcb.fDtrControl = DTR_CONTROL_ENABLE;

  // Set the baud rate
  switch (baud) {
    case    300: dcb.BaudRate = CBR_300;    break;
    case   1200: dcb.BaudRate = CBR_1200;   break;
    case   2400: dcb.BaudRate = CBR_2400;   break;
    case   4800: dcb.BaudRate = CBR_4800;   break;
    case   9600: dcb.BaudRate = CBR_9600;   break;
    case  19200: dcb.BaudRate = CBR_19200;  break;
    case  38400: dcb.BaudRate = CBR_38400;  break;
    case  57600: dcb.BaudRate = CBR_57600;  break;
    case 115200: dcb.BaudRate = CBR_115200; break;
    default:
      fprintf(stderr,"rs232_comm_open: unknown baud rate %ld\n",baud);
      CloseHandle(hCom);
      return -1;
  }

  /* no parity */
  dcb.fParity = FALSE;
  dcb.Parity = NOPARITY;
  dcb.StopBits = ONESTOPBIT;
  dcb.ByteSize = 8;

  if (!(fSuccess = SetCommState(hCom, &dcb))) {
    perror("rs232_comm_open: cannot set serial port configuration settings");
    CloseHandle(hCom);
    return -1;
  }
 
  cto.ReadIntervalTimeout = MAXDWORD;
  cto.ReadTotalTimeoutMultiplier = MAXDWORD;
  cto.ReadTotalTimeoutConstant = 1;
  cto.WriteTotalTimeoutConstant = 0;
  cto.WriteTotalTimeoutMultiplier = 0;

  if (!(fSuccess = SetCommTimeouts(hCom, &cto))) {
    perror("rs232_comm_open: cannot set serial port timeouts");
    CloseHandle(hCom);
    return -1;
  }

  *commhandleptr = NULL;
  comm = (commstruct *) malloc(sizeof(commstruct));
  if (comm == NULL)
    return -1;

  comm->hCom = hCom; /* set handle for later on */

  *commhandleptr = (RS232CommHandle) comm;

  return 0;
}


int rs232_comm_write(RS232CommHandle commhandle, const char * buf) {
  commstruct * comm = (commstruct *) commhandle;
  BOOL fSuccess;
  DWORD numWritten;
  OVERLAPPED Overlapped;

  if (comm == NULL)
    return -1;

  Overlapped.Offset = 0;
  Overlapped.OffsetHigh = 0;
  Overlapped.hEvent = NULL;

  if(!(fSuccess = WriteFile(comm->hCom, buf, strlen(buf), &numWritten, &Overlapped))) {
    perror("rs232_comm_write: can't write to serial port");
    return -1;
  }

  return numWritten;
}


int rs232_comm_read(RS232CommHandle commhandle, char * buf, int sz) {
  commstruct * comm = (commstruct *) commhandle;
  BOOL fSuccess;
  DWORD numRead;
  COMSTAT cstat;
  DWORD errors;
  OVERLAPPED Overlapped;

  if (comm == NULL)
    return -1;

  Overlapped.Offset = 0;
  Overlapped.OffsetHigh = 0;
  Overlapped.hEvent = NULL;

  if (!(fSuccess = ClearCommError(comm->hCom, &errors, &cstat))) {
    perror("rs232_comm_read: can't get current status");
    return -1;
  }

  if (cstat.cbInQue > 0) {
    if(!(fSuccess = ReadFile(comm->hCom, buf, sz, &numRead, &Overlapped))) {
      perror("rs232_comm_read: can't read from serial port");
      return -1;
    }

    return numRead;
  }

  return 0;
}


int rs232_comm_close(RS232CommHandle * commhandleptr) {
  commstruct * comm = (commstruct *) (*commhandleptr);

  if (comm == NULL)
    return -1;

  CloseHandle(comm->hCom);

  free(*commhandleptr);
  *commhandleptr = NULL;

  return 0;
}

int rs232_comm_get_status(RS232CommHandle handle, int *dsr, int *dtr, int *rts, int *cts, int *cd, int *rng) {
  return -1; /* not implemented yet */
}

int rs232_comm_set_dtr(RS232CommHandle handle, int level) {
  return -1; /* not implemented yet */
}

int rs232_comm_set_rts(RS232CommHandle handle, int level) {
  return -1; /* not implemented yet */
}

#endif 


